package certinspector

import (
	"crypto/x509"
	"encoding/json"
	"encoding/pem"
	"errors"
	"fmt"
	"log"
	"os"
	"path/filepath"
	"sort"
	"strings"
	"sync"
	"time"
)

var (
	DebugLogger   *log.Logger
	InfoLogger    *log.Logger
	WarningLogger *log.Logger
	ErrorLogger   *log.Logger
)

func init() {
	DebugLogger = log.New(os.Stdout, "DEBUG: ", log.Ldate|log.Ltime)
	InfoLogger = log.New(os.Stdout, "INFO: ", log.Ldate|log.Ltime)
	WarningLogger = log.New(os.Stdout, "WARNING: ", log.Ldate|log.Ltime)
	ErrorLogger = log.New(os.Stdout, "ERROR: ", log.Ldate|log.Ltime)
}

type CertificateData struct {
	CertName     string   `json:"certname"`
	SerialNumber string   `json:"serial_number"`
	Domains      []string `json:"domains"`
	ExpiryDate   string   `json:"expiry_date"`
}

// Inspect certificate and return CertificateData.
func InspectCertificate(certDir string, verboseBool bool) (CertificateData, error) {
	if verboseBool {
		DebugLogger.Printf("executing InspectCertificate function for: %s", certDir)
	}
	fullchainPath := filepath.Join(certDir, "fullchain.pem")
	data, err := os.ReadFile(fullchainPath)
	if err != nil {
		return CertificateData{}, fmt.Errorf("failed to read certificate: %w", err)
	}

	block, _ := pem.Decode(data)
	if block == nil {
		return CertificateData{}, errors.New("failed to parse PEM block")
	}

	cert, err := x509.ParseCertificate(block.Bytes)
	if err != nil {
		return CertificateData{}, fmt.Errorf("failed to parse certificate: %w", err)
	}

	certName := cert.Subject.CommonName
	serial := strings.ToUpper(fmt.Sprintf("%X", cert.SerialNumber))
	domains := cert.DNSNames
	daysLeft := int(time.Until(cert.NotAfter).Hours() / 24)
	status := "VALID"
	if daysLeft <= 0 {
		status = "EXPIRED"
	}
	expiry := fmt.Sprintf("%s: %d DAYS", status, daysLeft)

	return CertificateData{
		CertName:     certName,
		SerialNumber: serial,
		Domains:      domains,
		ExpiryDate:   expiry,
	}, nil
}

// Process certificate and call writeJSON function. Write to file.
func ProcessCertificatesWrite(baseDir, provider string, outputDir string, verboseBool bool) error {
	if verboseBool {
		DebugLogger.Printf("executing ProcessCertificatesWrite function for provider: %s", provider)
	}
	liveDir := filepath.Join(baseDir, provider, "live")
	dirs, err := os.ReadDir(liveDir)
	if err != nil {
		return fmt.Errorf("failed to list directories: %w", err)
	}

	var wg sync.WaitGroup
	var mu sync.Mutex
	results := []CertificateData{}

	for _, dir := range dirs {
		if !dir.IsDir() || dir.Name() == "README" {
			continue
		}

		wg.Add(1)
		go func(certDir string) {
			defer wg.Done()
			data, err := InspectCertificate(certDir, verboseBool)
			if err == nil {
				mu.Lock()
				results = append(results, data)
				mu.Unlock()
			}
		}(filepath.Join(liveDir, dir.Name()))
	}

	wg.Wait()

	sort.Slice(results, func(i, j int) bool {
		return results[i].CertName < results[j].CertName
	})

	outputFile := filepath.Join(outputDir, provider+".json")
	return writeJSON(outputFile, results, verboseBool)
}

// process certificates and return JSON data. Used by the API. It doesn't write to file.
func ProcessCertificates(baseDir, provider string, verboseBool bool) ([]byte, error) {
	if verboseBool {
		DebugLogger.Printf("executing ProcessCertificates for provider: %s", provider)
	}
	liveDir := filepath.Join(baseDir, provider, "live")
	dirs, err := os.ReadDir(liveDir)
	if err != nil {
		return nil, fmt.Errorf("failed to list directories: %w", err)
	}

	var wg sync.WaitGroup
	var mu sync.Mutex
	results := []CertificateData{}

	for _, dir := range dirs {
		if !dir.IsDir() || dir.Name() == "README" {
			continue
		}

		wg.Add(1)
		go func(certDir string) {
			defer wg.Done()
			data, err := InspectCertificate(certDir, verboseBool)
			if err == nil {
				mu.Lock()
				results = append(results, data)
				mu.Unlock()
			}
		}(filepath.Join(liveDir, dir.Name()))
	}

	wg.Wait()

	sort.Slice(results, func(i, j int) bool {
		return results[i].CertName < results[j].CertName
	})

	jsonData, err := json.MarshalIndent(results, "", "  ")
	if err != nil {
		return nil, fmt.Errorf("failed to encode JSON: %w", err)
	}

	return jsonData, nil
}

// write JSON to file. Used by the ProcessCertificatesWrite function.
func writeJSON(filename string, data interface{}, verboseBool bool) error {
	if verboseBool {
		DebugLogger.Printf("executing writeJSON function, target file: %s", filename)
	}
	file, err := os.Create(filename)
	if err != nil {
		return fmt.Errorf("failed to create JSON file: %w", err)
	}
	defer file.Close()

	encoder := json.NewEncoder(file)
	encoder.SetIndent("", "  ")
	return encoder.Encode(data)
}
